from functools import partial

import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from .averaging import compute_grands, compute_grands_df
from .io import (besa_extensions, convert_participant_input, eeg_extensions,
                 files_from_dir, get_participant_id, log_extensions,
                 package_versions, save_config, save_df, save_evokeds)
from .participant import participant_pipeline
from .perm import compute_perm, compute_perm_tfr


def group_pipeline(
    raw_files=None,
    log_files=None,
    output_dir=None,
    clean_dir=None,
    epochs_dir=None,
    report_dir=None,
    to_df=True,
    downsample_sfreq=None,
    veog_channels='auto',
    heog_channels='auto',
    montage='easycap-M1',
    bad_channels=None,
    ref_channels='average',
    besa_files=None,
    ica_method=None,
    ica_n_components=None,
    highpass_freq=0.1,
    lowpass_freq=40.0,
    triggers=None,
    triggers_column=None,
    epochs_tmin=-0.5,
    epochs_tmax=1.5,
    baseline=(-0.2, 0.0),
    skip_log_rows=None,
    skip_log_conditions=None,
    reject_peak_to_peak=200.0,
    components={'name': [], 'tmin': [], 'tmax': [], 'roi': []},
    average_by=None,
    perform_tfr=False,
    tfr_subtract_evoked=False,
    tfr_freqs=np.linspace(4.0, 40.0, num=37),
    tfr_cycles=np.linspace(2.0, 20.0, num=37),
    tfr_mode='percent',
    tfr_baseline=(-0.45, -0.05),
    tfr_components={
        'name': [], 'tmin': [], 'tmax': [], 'fmin': [], 'fmax': [], 'roi': []},
    perm_contrasts=[],
    perm_tmin=0.0,
    perm_tmax=1.0,
    perm_channels=None,
    perm_fmin=None,
    perm_fmax=None,
    n_jobs=1,
    vhdr_files=None
):
    """Process EEG data for a group of participants.

    Performs preprocessing and computes single trial mean amplitudes for ERP
    components of interest as well as by-participant averaged waveforms.
    Optionally, performs time-frequency analysis and/or cluster-based
    permutation tests.

    Parameters & returns
    --------------------
    See `Usage <../usage.html>`_ for the pipeline input arguments and outputs.
    """

    # Convert input types
    tfr_freqs = list(tfr_freqs)
    tfr_cycles = list(tfr_cycles)

    # Backup input arguments for re-use
    config = locals().copy()

    # Create partial function with arguments shared across participants
    partial_pipeline = partial(
        participant_pipeline,
        skip_log_conditions=skip_log_conditions,
        downsample_sfreq=downsample_sfreq,
        veog_channels=veog_channels,
        heog_channels=heog_channels,
        montage=montage,
        ref_channels=ref_channels,
        ica_method=ica_method,
        ica_n_components=ica_n_components,
        highpass_freq=highpass_freq,
        lowpass_freq=lowpass_freq,
        triggers=triggers,
        triggers_column=triggers_column,
        epochs_tmin=epochs_tmin,
        epochs_tmax=epochs_tmax,
        baseline=baseline,
        reject_peak_to_peak=reject_peak_to_peak,
        components=components,
        average_by=average_by,
        perform_tfr=perform_tfr,
        tfr_subtract_evoked=tfr_subtract_evoked,
        tfr_freqs=tfr_freqs,
        tfr_cycles=tfr_cycles,
        tfr_mode=tfr_mode,
        tfr_baseline=tfr_baseline,
        tfr_components=tfr_components,
        clean_dir=clean_dir,
        epochs_dir=epochs_dir,
        chanlocs_dir=output_dir,
        report_dir=report_dir,
        to_df=to_df)

    if raw_files is None:
        if vhdr_files is not None:
            from warnings import warn
            warn('⚠️ The `vhdr_files` argument has been renamed to `raw_files` ' +
                 'and will cease to work in a future version of the pipeline. ' +
                 'Please update your code accordingly.')
            raw_files = vhdr_files

    if ica_method is not None and ica_n_components is None:
        from warnings import warn
        warn('The default value of `ica_n_components` has changed from ' +
             '`0.99` (i.e., 99% explained variance) to `None` (i.e., ' +
             'extract as many components as possible). To reproduce ' +
             'previous results, explicitly set `ica_n_components=0.99`.')

    # Get input file paths if directories were provided
    if isinstance(raw_files, str):
        raw_files = files_from_dir(raw_files, eeg_extensions)
    if isinstance(log_files, str):
        log_files = files_from_dir(log_files, log_extensions)
    assert len(log_files) == len(raw_files), \
        f'Number of `log_files` ({len(log_files)}) does not match ' + \
        f'number of `raw_files` ({len(raw_files)})'

    # Get input BESA matrix files if necessary
    if isinstance(besa_files, str):
        besa_files = files_from_dir(besa_files, besa_extensions)
    elif besa_files is None:
        besa_files = [None] * len(raw_files)
    assert len(besa_files) == len(raw_files), \
        f'Number of `besa_files` ({len(besa_files)}) does not match ' + \
        f'number of `raw_files` ({len(raw_files)})'

    # Extract participant IDs from filenames
    participant_ids = [get_participant_id(f) for f in raw_files]

    # Construct lists of bad_channels and skip_log_rows per participant
    bad_channels = convert_participant_input(bad_channels, participant_ids)
    skip_log_rows = convert_participant_input(skip_log_rows, participant_ids)

    # Combine participant-specific inputs
    participant_args = zip(raw_files, log_files, besa_files,
                           bad_channels, skip_log_rows)

    # Do processing in parallel
    n_jobs = int(n_jobs)
    res = Parallel(n_jobs)(
        delayed(partial_pipeline)(*args) for args in participant_args)

    # Sort outputs into seperate lists
    print(f'\n\n=== Processing group level ===')
    trials, evokeds, evokeds_dfs, configs = list(map(list, zip(*res)))[0:4]

    # Combine trials and save
    trials = pd.concat(trials, ignore_index=True)
    save_df(trials, output_dir, suffix='trials')

    # Combine evokeds_dfs and save
    evokeds_df = pd.concat(evokeds_dfs, ignore_index=True)
    save_df(evokeds_df, output_dir, suffix='ave')

    # Compute grand averaged ERPs and save
    grands = compute_grands(evokeds)
    grands_df = compute_grands_df(evokeds_df)
    save_evokeds(
        grands, grands_df, output_dir, participant_id='grand', to_df=to_df)

    # Update config with participant-specific inputs...
    config['raw_files'] = raw_files
    config['bad_channels'] = bad_channels
    config['besa_files'] = besa_files
    config['skip_log_rows'] = skip_log_rows

    # ... and outputs that might have been created along the way
    config['log_files'] = [p_config['log_file'] for p_config in configs]
    auto_keys = ['auto_bad_channels', 'auto_missing_epochs',
                 'auto_rejected_epochs', 'auto_ica_n_components',
                 'auto_ica_bad_components']
    for key in auto_keys:
        if any(key in p_config and p_config[key] is not None
               for p_config in configs):
            config[key] = {p_id: p_config[key] for p_id, p_config
                           in zip(participant_ids, configs)}

    # Save config
    config['package_versions'] = package_versions()
    save_config(config, output_dir)

    # Define standard returns
    returns = [trials, evokeds_df, config]

    # Cluster based permutation tests for ERPs
    if perm_contrasts != []:
        cluster_df = compute_perm(evokeds, perm_contrasts, perm_tmin,
                                  perm_tmax, perm_channels, n_jobs)
        save_df(cluster_df, output_dir, suffix='clusters')
        returns.append(cluster_df)

    # Combine time-frequency results
    if perform_tfr:

        # Sort outputs into seperate lists
        tfr_evokeds, tfr_evokeds_dfs = list(map(list, zip(*res)))[4:6]

        # Combine evokeds_df for power and save
        tfr_evokeds_df = pd.concat(tfr_evokeds_dfs, ignore_index=True)
        save_df(tfr_evokeds_df, output_dir,
                suffix='tfr_ave')
        returns.append(tfr_evokeds_df)

        # Compute grand averaged power and save
        tfr_grands = compute_grands(tfr_evokeds)
        tfr_grands_df = compute_grands_df(tfr_evokeds_df)
        save_evokeds(tfr_grands, tfr_grands_df, output_dir,
                     participant_id='tfr_grand', to_df=to_df)

        # Cluster based permutation tests for TFR
        if perm_contrasts != []:
            tfr_cluster_df = compute_perm_tfr(
                tfr_evokeds, perm_contrasts, perm_tmin, perm_tmax,
                perm_channels, perm_fmin, perm_fmax, n_jobs)
            save_df(tfr_cluster_df, output_dir, suffix='tfr_clusters')
            returns.append(tfr_cluster_df)

    return returns



from warnings import warn

import pandas as pd
import pooch
from numpy import nan
from pandas.api.types import is_list_like

LOCAL_CACHE = 'hu-neuro-pipeline'


def get_dataset(manifest_df, base_url, participants, path):

    if path is None:
        path = pooch.os_cache(LOCAL_CACHE)
        env = 'PIPELINE_DATA_DIR'
    else:
        env = None

    fetcher = pooch.create(path=path, base_url=base_url, env=env)
    local_dir = fetcher.abspath

    manifest_df = _select_participants(manifest_df, participants)

    file_types = manifest_df['file_type'].unique()
    file_types = file_types[~pd.isnull(file_types)]
    paths = {file_type: [] for file_type in file_types}

    for ix, row in manifest_df.iterrows():

        local_file = local_dir.joinpath(row['local_path'])

        if not local_file.exists():
            fetcher.registry[row['local_path']] = row['hash']
            fetcher.urls[row['local_path']] = row['url']
            _ = fetcher.fetch(row['local_path'])

        if row['file_type'] in paths:
            paths[row['file_type']].append(str(local_file))

    return paths


def _select_participants(df, participants):
    """Selects a subset of participants by their IDs or total number."""

    all_participants = df['participant_id'].str.zfill(2).unique().tolist()
    if nan in all_participants:  # Ignore general (e.g., BIDS) files for now
        all_participants.remove(nan)

    if isinstance(participants, float):
        warn(f'Converting `participants` from float ({participants}) to ' +
             f'int ({int(participants)})')
        selected_participants = all_participants[:int(participants)]

    if isinstance(participants, int):
        assert participants in range(1, len(all_participants) + 1), \
            '`participants` must be an integer between 1 and ' + \
            f'{len(all_participants)}'
        selected_participants = all_participants[:participants]

    if isinstance(participants, str):
        assert participants in all_participants, \
            f'Participant \'{participants}\' not found in the dataset. ' + \
            f'Valid participants are {all_participants}'
        selected_participants = [participants]

    if is_list_like(participants):
        missing_participants = list(set(participants) - set(all_participants))
        assert not missing_participants, \
            f'Participants {missing_participants} not found in the ' + \
            f'dataset. Valid participants are {all_participants}'
        selected_participants = participants

    selected_participants += [nan]  # Re-include general (e.g., BIDS) files

    return df[df['participant_id'].isin(selected_participants)]



import json
from pathlib import Path
from urllib.request import urlopen

import pandas as pd

from .utils import get_dataset

BASE_URL = 'https://files.de-1.osf.io/v1/resources/hdxvb/providers/osfstorage'
MANIFEST_FILE = Path(__file__).parent.joinpath('ucap_manifest.csv')

FILE_TYPE_DICT = {'vhdr': 'raw_files',
                  'txt': 'log_files',
                  'matrix': 'besa_files'}


def get_ucap(participants=40, path=None):
    """Get sample data from the UCAP dataset.

    Data that are not yet available locally will be downloaded from the OSF.
    See :footcite:`fromer2018` for details on the UCAP dataset.

    Parameters
    ----------
    participants : int or list of str, optional
        Which participants to download. By default, downloads all 40
        participants available in the dataset. If an integer, downloads that
        many participants starting from the first participant. If a list of
        strings, downloads the participants with the given IDs (e.g.,
        ``['05', '07']``).
    path : str or Path, optional
        Local directory path to download the data to. By default, uses the
        user's local cache directory. An alternative way to specify the
        download path is to set the environment variable ``PIPELINE_DATA_DIR``.

    Returns
    -------
    dict
        A dictionary with the file paths of the downloaded data:

        - ``'raw_files'``: A list with the paths of the raw EEG files
          (``.vhdr``)
        - ``'log_files'`` A list with the paths of the log files (``.txt``)
        - ``'besa_files'`` A list with the paths of the BESA calibration files
          (``.matrix``)

    See Also
    --------
    pipeline.datasets.get_erpcore

    References
    ----------
    .. footbibliography::
    """

    manifest_df = pd.read_csv(MANIFEST_FILE, dtype={'participant_id': str})

    return get_dataset(manifest_df, BASE_URL, participants, path)


def _write_ucap_manifest():
    """Writes a CSV table containing the file paths of the UCAP dataset."""

    eeg_url = '59cf07fa6c613b02958f3364/'
    log_url = '59cf12259ad5a102cc5c4b93/'
    cali_url = '59cf089e6c613b02968f5724/'

    files = []
    for url in [eeg_url, log_url, cali_url]:
        with urlopen(f'{BASE_URL}/{url}') as url:
            files += json.loads(url.read().decode())['data']

    attributes = [file['attributes'] for file in files]

    df = pd.DataFrame.from_dict(attributes)

    participants = df['name'].str.split('_|\.').str[0].str.zfill(2)

    n_expected_files = 5  # Complete participants have 3 x EEG, 1 x log, 1 x cali
    n_files = participants.value_counts()
    good_participant_ids = n_files[n_files == n_expected_files].index.to_list()

    df.insert(0, 'participant_id', participants)
    df = df.sort_values(by=['participant_id', 'name'])
    df = df[df['participant_id'].isin(good_participant_ids)]

    local_paths = df['materialized'].str.replace('/UCAP/Data/', 'ucap/')
    df.insert(1, 'local_path', local_paths)

    hashes = df['extra'].apply(lambda x: f'md5:{x["hashes"]["md5"]}')
    df.insert(2, 'hash', hashes)

    urls = df['path'].apply(lambda x: f'{BASE_URL}{x}')
    df.insert(3, 'url', urls)

    file_exts = df['name'].apply(lambda x: Path(x).suffix[1:])
    file_types = file_exts.map(FILE_TYPE_DICT)
    df.insert(4, 'file_type', file_types)

    df = df[['local_path', 'url', 'hash', 'participant_id',
             'file_type', 'size']]

    df.to_csv(MANIFEST_FILE, index=False)



import json
from pathlib import Path
from urllib.request import urlopen

import pandas as pd

from .utils import get_dataset

LOCAL_CACHE = 'hu-neuro-pipeline/erpcore'
MANIFEST_FILE = Path(__file__).parent.joinpath('erpcore_manifest.csv')

OSF_IDS = {'ERN': 'q6gwp',
           'LRP': '28e6c',
           'MMN': '5q4xs',
           'N170': 'pfde9',
           'N2pc': 'yefrq',
           'N400': '29xpq',
           'P3': 'etdkz'}

FILE_TYPE_DICT = {'eeg.set': 'raw_files',
                  'events.tsv': 'log_files'}


def get_erpcore(component, participants=40, path=None):
    """Get sample data from the ERP CORE dataset.

    Data that are not yet available locally will be downloaded from the OSF.
    See :footcite:`kappenman2021` for details on the ERP CORE dataset.

    Parameters
    ----------
    component : str
        Which ERP CORE experiment to download. Each experiment was designed to
        elicit one of seven common ERP components:

        - ``'ERN'`` (flanker task)
        - ``'LRP'`` (flanker task)
        - ``'MMN'`` (passive auditory oddball task)
        - ``'N170'`` (face perception task)
        - ``'N2pc'`` (simple visual search task)
        - ``'N400'`` (word pair judgment task)
        - ``'P3'`` (active visual oddball task)

    participants : int or list of str, optional
        Which participants to download. By default, downloads all 40
        participants available in the dataset. If an integer, downloads that
        many participants starting from the first participant. If a list of
        strings, downloads the participants with the given IDs (e.g.,
        ``['sub-001', 'sub-002']``).
    path : str or Path, optional
        Local directory path to download the data to. By default, uses the
        user's local cache directory. An alternative way to specify the
        download path is to set the environment variable ``PIPELINE_DATA_DIR``.

    Returns
    -------
    dict
        A dictionary with the file paths of the downloaded data:

        - ``'raw_files'``: A list with the paths of the raw EEG files
          (``eeg.set``)
        - ``'log_files'`` A list with the paths of the log files
          (``events.tsv``)

    See Also
    --------
    pipeline.datasets.get_ucap

    References
    ----------
    .. footbibliography::
    """

    manifest_df = pd.read_csv(MANIFEST_FILE, dtype={'participant_id': str})
    manifest_df = manifest_df[manifest_df['component'] == component]

    osf_id = OSF_IDS[component]
    base_url = f'https://files.de-1.osf.io/v1/resources/{osf_id}/providers/osfstorage'

    return get_dataset(manifest_df, base_url, participants, path)


def _write_erpcore_manifest():
    """Writes a CSV table containing the file paths of the ERP CORE datasets."""

    dfs = []
    for component, osf_id in OSF_IDS.items():

        base_url = f'https://files.de-1.osf.io/v1/resources/{osf_id}/providers/osfstorage/'

        bids_suffix = _find_bids_remote_path(base_url)

        files = _list_files(base_url, bids_suffix, exclude_dirs=['stimuli'])

        attributes = [file['attributes'] for file in files]
        df = pd.DataFrame.from_dict(attributes)

        df.insert(0, 'component', component)

        participants = df['name'].str.split('_|\.').str[0]
        participants = [p if p.startswith('sub') else '' for p in participants]
        df.insert(1, 'participant_id', participants)
        df = df.sort_values('participant_id')

        local_paths = df['materialized'].str.\
            replace(f'/{component} Raw Data BIDS-Compatible/',
                    f'erpcore/{component}/')
        df.insert(2, 'local_path', local_paths)

        hashes = df['extra'].apply(lambda x: f'md5:{x["hashes"]["md5"]}')
        df.insert(3, 'hash', hashes)

        urls = df['path'].apply(lambda x: f'{base_url}{x}')
        df.insert(4, 'url', urls)

        file_exts = df['name'].str.split('_').str[-1]
        file_types = file_exts.map(FILE_TYPE_DICT)
        df.insert(5, 'file_type', file_types)

        df = df[['component', 'local_path', 'url', 'hash', 'participant_id',
                 'file_type', 'size']]

        dfs.append(df)

    df = pd.concat(dfs, ignore_index=True)

    df.to_csv(MANIFEST_FILE, index=False)


def _find_bids_remote_path(base_url):
    """Finds the BIDS directory for a given ERP CORE component dataaset."""

    with urlopen(base_url) as url:
        files = json.loads(url.read().decode())['data']
        bids_dir = [f for f in files
                    if 'Raw Data BIDS-Compatible'
                    in f['attributes']['name']][0]

    return bids_dir['attributes']['path']


def _list_files(base_url, suffix, exclude_dirs=None):
    """Lists files recursively in a remote directory on OSF."""

    with urlopen(f'{base_url}/{suffix}') as url:
        files = json.loads(url.read().decode())['data']

        for file in files:
            if file['attributes']['kind'] == 'folder':

                if exclude_dirs is not None \
                        and file['attributes']['name'] in exclude_dirs:
                    continue

                new_suffix = file['attributes']['path']
                files += _list_files(base_url, new_suffix, exclude_dirs)

        files = [file for file in files
                 if file['attributes']['kind'] == 'file']

    return files



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import inspect
import os
import sys
from pathlib import Path

# Make sure the pipeline package is available
sys.path.insert(0, Path(__file__).parents[2].resolve().as_posix())
import pipeline

# Make sure Quarto and its dependencies are available
# This seems to be necessary when install Quarto via conda -- it doesn't by
# itself find the `share` directory or `deno` in the correct places
bin_path = Path(sys.executable).parent
share_path = bin_path.parent.joinpath('share')
os.environ['QUARTO_SHARE_PATH'] = share_path.joinpath('quarto').resolve().as_posix()
os.environ['DENO_DIR'] = bin_path.resolve().as_posix()
os.environ['DENO_BIN'] = bin_path.joinpath('deno').resolve().as_posix()
os.environ['QUARTO_DENO'] = bin_path.joinpath('deno').resolve().as_posix()

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'hu-neuro-pipeline'
copyright = '2023, Alexander Enge'
author = 'Alexander Enge'
version = '.'.join(pipeline.__version__.split('.', 2)[:2])
release = pipeline.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.intersphinx',
              'sphinx.ext.linkcode',
              'sphinx.ext.napoleon',
              'sphinxcontrib.bibtex',
              'sphinxcontrib.apa',
              'myst_nb',
              'sphinx_copybutton',
              'sphinx_gallery.load_style']
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']
source_suffix = {
    '.rst': 'restructuredtext',
    '.qmd': 'myst-nb'
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_title = 'hu-neuro-pipeline'
html_logo = '_static/logo.png'
html_favicon = '_static/favicon.png'
html_theme = 'sphinx_book_theme'
html_theme_options = {
    'path_to_docs': 'doc',
    'repository_url': 'https://github.com/alexenge/hu-neuro-pipeline',
    'repository_branch': 'main',
    'use_repository_button': True,
    'use_issues_button': True,
    'use_edit_page_button': True,
    'use_fullscreen_button': False,
    'extra_navbar': ''}
html_static_path = ['_static']
html_css_files = ['custom.css']
pygments_style = 'tango'

# -- Options for sphinx.linkscode --------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/linkcode.html

def linkcode_resolve(domain, info):
    def find_source():
        # try to find the file and line number, based on code from numpy:
        # https://github.com/numpy/numpy/blob/master/doc/source/conf.py#L286
        obj = sys.modules[info['module']]
        for part in info['fullname'].split('.'):
            obj = getattr(obj, part)
        fn = inspect.getsourcefile(obj)
        fn = os.path.relpath(fn, start=os.path.dirname(pipeline.__file__))
        source, lineno = inspect.getsourcelines(obj)
        return fn, lineno, lineno + len(source) - 1

    if domain != 'py' or not info['module']:
        return None
    try:
        filename = 'pipeline/%s#L%d-L%d' % find_source()
    except Exception:
        filename = info['module'].replace('.', '/') + '.py'
    tag = 'main' if 'dev' in release else ('v' + release)

    return "https://github.com/alexenge/hu-neuro-pipeline/blob/%s/%s" % (tag, filename)

# -- InterSphinx options -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'mne': ('https://mne.tools/stable', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'scipy': ('https://docs.scipy.org/doc/scipy', None),
    'matplotlib': ('https://matplotlib.org/stable', None),
    'sklearn': ('https://scikit-learn.org/stable', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable', None),
}

# -- Napoleon options --------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html#configuration

napoleon_preprocess_types = True

## -- sphinxcontrib-bibtex options -------------------------------------------
# https://sphinxcontrib-bibtex.readthedocs.io/en/latest/usage.html#configuration

bibtex_bibfiles = ['references.bib']
bibtex_default_style = 'apa'

# -- nbsphinx options --------------------------------------------------------
# https://nbsphinx.readthedocs.io/en/latest/configuration.html

nb_execution_timeout = 600
nb_custom_formats = {
    '.pct.py': ['jupytext.reads', {'fmt': 'py:percent'}],
    '.qmd': ['jupytext.reads', {'fmt': 'quarto'}],
    '.Rmd': ['jupytext.reads', {'fmt': 'Rmd'}]
}
nb_render_image_options = {'width': '70%', 'align': 'center'}
